import { select } from "d3-selection";
import "d3-transition";
import { getTypeFromChartLayout, remToPx } from "./utils";
import {min, max} from "d3-array";
import { wrapStringToLines } from "@flourish/pocket-knife";

function initUpdate(instance, state) {
	instance.update = function() {
		const chart_layout = instance.chartLayout();
		const facet_name = instance.facetName();
		const line_height = 1.20;
		const data = [{ chart_layout, facet_name: facet_name}];

		const fg_class = "fg-label";
		const bg_class = "bg-label";


		var parseNum = instance.numberParser();
		var dur = instance.animationDuration();

		function getXTextSize() {
			return remToPx(state.x_font_size);
		}

		function getYTextSize() {
			return remToPx(state.y_font_size);
		}

		function splitLabelAndValue(string) {
			return string.replace(/\s*::\s*/g, "::").split("::").map(s => s.trim());
		}

		function splitRange(string) {
			return string.replace(/\s*>>\s*/, ">>").split(">>").map(s => s.trim());
		}

		function getColor(d) {
			return d.color;
		}

		function getParsedValue(value, type, chart_layout, axis) {
			let parsed_value;
			if (type === "number") parsed_value = isNaN(parseNum(value)) ? parseNum(value) : value;
			else if (type === "datetime") {
				const parser = axis === "x" ? chart_layout.xDatetimeParse() : chart_layout.yDatetimeParse();
				parsed_value = parser(value);
			}
			else parsed_value = value;
			return parsed_value;
		}

		function updateXLines(facet_group, data) {
			const chart_layout = data.chart_layout;
			let margin_left = chart_layout.margins.left;
			let margin_top = chart_layout.margins.top;
			let plot_height = Math.max(0, chart_layout.plot_height);
			let disabled = instance.xDisabled();

			let xScale = chart_layout.xScale();
			let type = getTypeFromChartLayout(chart_layout.xData());

			var line_data = [];

			if (!disabled && state.x_enabled && state.x_lines) {
				state.x_lines.split("\n").forEach(function(string) {
					var line = splitLabelAndValue(string);
					var label = line[0];
					var value = line[1];
					if (value === undefined) return;
					var color = line[2] || state.x_line_color;
					var facet = line[3] || null;
					if (facet && data.facet_name !== facet) return;
					line_data.push({
						label: label,
						x: getParsedValue(value, type, chart_layout, "x"),
						color: color
					});
				});
			}

			var lines = select(facet_group).selectAll("g.line.x").data(line_data);

			var lines_enter = lines.enter().append("g").attr("class", "line x")
				.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; });
			lines_enter.append("line");
			lines_enter.append("text");

			var lines_update = lines.merge(lines_enter);
			lines_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; });
			lines_update.select("line")
				.attr("stroke-dasharray", state.x_line_dash + " " + state.x_line_dash)
				.attr("stroke", getColor)
				.attr("stroke-width", state.x_line_width)
				.attr("y2", plot_height);
			lines_update.select("text").each(function(d) {
				const x_coordinate = (xScale(d.x) || margin_left);
				const font_family = instance.fontFamily();
				const max_width_in_px = remToPx(state.x_label_max_width);
				const font_style = `${state.x_font_weight}  ${state.x_font_size} ${font_family}`;
				const wrapped_lines_array = wrapStringToLines(d.label,
					font_style,
					null,
					max_width_in_px);


				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, bg_class);
				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, fg_class);

				const num_of_lines = wrapped_lines_array.length;
				const {transform_str, text_anchor} = getTransformAndAnchorForX(this, xScale, plot_height, num_of_lines, x_coordinate);
				select(this)
					.attr("transform", transform_str)
					.attr("text-anchor", text_anchor);
			})
			.attr("font-size", getXTextSize());

			lines_update.selectAll(`.${fg_class}`)
				.attr("fill", state.x_label_color);

			lines_update.selectAll(`.${bg_class}`)
				.attr("display", state.x_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.x_shadow_color)
				.attr("stroke-width", 2);


			lines.exit().remove();
		}

		function updateXAreas(facet_group, data) {
			const chart_layout = data.chart_layout;

			let margin_left = chart_layout.margins.left;
			let margin_top = chart_layout.margins.top;
			let plot_height = Math.max(0, chart_layout.plot_height);
			let plot_width = Math.max(0, chart_layout.plot_width);
			let xScale = chart_layout.xScale();
			let type = getTypeFromChartLayout(data.chart_layout.xData());
			let disabled = instance.xDisabled();

			var area_data = [];
			if (!disabled && state.x_enabled && state.x_areas) {
				state.x_areas.split("\n").map(function(string) {
					var area = splitLabelAndValue(string);
					if (area.length < 2) return;
					var label = area[0];
					var range = splitRange(area[1]).map(function(value, i) {
						var x_val = xScale(getParsedValue(value, type, chart_layout, "x"));
						if (instance.xExtendByStep()) x_val += xScale.step()/2 * (i == 0 ? -1 : 1);
						return x_val;
					});
					var color = area[2] || state.x_area_color;
					var facet = area[3] || null;
					if (facet && data.facet_name !== facet) return;
					if (range.length > 1) area_data.push({ label: label, range: range, color: color });
				});
			}

			var areas = select(facet_group).selectAll("g.area.x").data(area_data);
			var areas_enter = areas.enter().append("g").attr("class", "area x")
				.attr("transform", function(d) { return "translate(" + (Math.min(d.range[0], d.range[1]) || margin_left) + ", " + margin_top + ")"; });

			areas_enter.append("rect").attr("height", plot_height)
				.attr("width", function(d) { return (Math.abs(d.range[1] - d.range[0])) || 0; });
			areas_enter.append("text");

			var areas_update = areas.merge(areas_enter);
			areas_update.transition().duration(dur)
				.attr("transform", function(d) {
					const min_val = min(d.range);
					const min_val_or_y_axis = Math.max(min_val, margin_left);
					return "translate(" + (min_val_or_y_axis || 0) + ", " + margin_top + ")";
				});
			areas_update.select("rect")
				.attr("width", function(d) {
					const min_val = min(d.range);
					const max_val = max(d.range);
					const min_val_or_y_axis = Math.max(min_val, margin_left);
					const end_of_chart = plot_width + margin_left;
					const max_val_or_end_of_chart = Math.min(max_val, end_of_chart);
					const width = Math.abs(max_val_or_end_of_chart - min_val_or_y_axis);

					return width || 0;
				})
				.attr("fill", getColor)
				.attr("opacity", state.x_area_opacity)
				.attr("height", plot_height);
			areas_update.select("text").each(function(d) {
				const font_family = instance.fontFamily();
				const x_coordinate = (Math.min(d.range[0], d.range[1]) || margin_left);
				const max_width_in_px = remToPx(state.x_label_max_width);
				const font_style = `${state.x_font_weight} ${state.x_font_size} ${font_family}`;

				const wrapped_lines_array = wrapStringToLines(d.label,
					font_style,
					null,
					max_width_in_px);
				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, bg_class);
				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, fg_class);

				const num_of_lines = wrapped_lines_array.length;
				const {transform_str, text_anchor} = getTransformAndAnchorForX(this, xScale, plot_height, num_of_lines, x_coordinate);
				select(this)
					.attr("transform", transform_str)
					.attr("text-anchor", text_anchor);
			})
			.attr("font-size", getXTextSize());

			areas_update.selectAll(`.${fg_class}`)
				.attr("fill", state.x_label_color);

			areas_update.selectAll(`.${bg_class}`)
				.attr("display", state.x_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.x_shadow_color)
				.attr("stroke-width", 2);


			areas.exit().remove();
		}

		function updateYLines(facet_group, data) {
			const chart_layout = data.chart_layout;
			let plot_width = Math.max(0, chart_layout.plot_width);
			let x = chart_layout.margins.left;
			let yScale = chart_layout.yScale();
			let type = getTypeFromChartLayout(chart_layout.yData());
			let disabled = instance.yDisabled();
			let margin_top = chart_layout.margins.top;

			var line_data = [];
			if (!disabled && state.y_enabled && state.y_lines) {
				state.y_lines.split("\n").forEach(function(string) {
					var line = splitLabelAndValue(string);
					var label = line[0];
					var value = line[1];
					if (value === undefined) return;
					var color = line[2] || state.y_line_color;
					var facet = line[3] || null;
					if (facet && data.facet_name !== facet) return;
					line_data.push({
						label: label,
						y: getParsedValue(value, type, chart_layout, "y"),
						color: color
					});
				});
			}

			var lines = select(facet_group).selectAll("g.line.y").data(line_data);

			var lines_enter = lines.enter().append("g").attr("class", "line y")
				.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; });
			lines_enter.append("line").attr("width", 0);
			lines_enter.append("text");

			var lines_update = lines.merge(lines_enter);
			lines_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; });
			lines_update.select("line")
				.attr("stroke-dasharray", state.y_line_dash + " " + state.y_line_dash)
				.attr("stroke", getColor)
				.attr("stroke-width", state.y_line_width)
				.attr("x2", plot_width);
			lines_update.select("text")
				.each(function(d) {
					const font_family = instance.fontFamily();
					const y_coordinate = (yScale(d.y) || margin_top);
					const max_width_in_px = remToPx(state.y_label_max_width);
					const font_style = `${state.y_font_weight} ${state.y_font_size} ${font_family}`;
					const wrapped_lines_array = wrapStringToLines(d.label,
						font_style,
						null,
						max_width_in_px);

					drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, bg_class);
					drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, fg_class);

					const num_of_lines = wrapped_lines_array.length;
					const {transform_str, text_anchor} = getTransformAndAnchorForY(this, yScale, plot_width, num_of_lines, y_coordinate);

					select(this)
						.attr("transform", transform_str)
						.attr("text-anchor", text_anchor);
				})
				.attr("font-size", getYTextSize());

			lines_update.selectAll(`.${fg_class}`)
				.attr("fill", state.y_label_color);

			lines_update.selectAll(`.${bg_class}`)
				.attr("display", state.y_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.y_shadow_color)
				.attr("stroke-width", 2);


			lines.exit().remove();
		}

		function updateYAreas(facet_group, data) {
			const chart_layout = data.chart_layout;
			let margin_top = chart_layout.margins.top;
			let plot_width = Math.max(0, chart_layout.plot_width);
			let plot_height = Math.max(0, chart_layout.plot_height);
			let x = chart_layout.margins.left;
			let yScale = chart_layout.yScale();
			let type = getTypeFromChartLayout(chart_layout.yData());
			let disabled = instance.yDisabled();

			var area_data = [];
			if (!disabled && state.y_enabled && state.y_areas) {
				state.y_areas.split("\n").map(function(string) {
					var area = splitLabelAndValue(string);
					if (area.length < 2) return;
					var label = area[0];
					var range = splitRange(area[1]).map(function(value, i) {
						var y_val = yScale(getParsedValue(value, type, chart_layout, "y"));
						if (instance.yExtendByStep()) y_val += yScale.step()/2 * (i == 0 ? 1 : -1);
						return y_val;
					});
					var color = area[2] || state.y_area_color;
					var facet = area[3] || null;
					if (facet && data.facet_name !== facet) return;
					if (range.length > 1) area_data.push({ label: label, range: range, color: color });
				});
			}

			var areas = select(facet_group).selectAll("g.area.y").data(area_data);

			var areas_enter = areas.enter().append("g").attr("class", "area y")
				.attr("transform", function(d) { return "translate(" + x + ", " + Math.min(d.range[0], d.range[1]) + ")"; });

			areas_enter.append("rect")
				.attr("width", plot_width)
				.attr("height", function(d) { return (Math.abs(d.range[1] - d.range[0])); });
			areas_enter.append("text");

			var areas_update = areas.merge(areas_enter);
			areas_update
				.transition().duration(dur)
				.attr("transform", function(d) {
					const min_val = min(d.range);
					const min_val_or_top_of_chart = Math.max(min_val, margin_top);
					return "translate(" + x + ", " + min_val_or_top_of_chart + ")";
				});
			areas_update.select("rect")
				.attr("height", function(d) {
					const min_val = min(d.range);
					const max_val = max(d.range);
					const bottom_of_chart = plot_height + margin_top;
					const max_val_or_bottom_of_chart = Math.min(max_val, bottom_of_chart);
					const min_val_or_top_of_chart = Math.max(min_val, margin_top);
					const height = Math.abs(max_val_or_bottom_of_chart - min_val_or_top_of_chart);
					return height;
				})
				.attr("fill", getColor)
				.attr("opacity", state.y_area_opacity)
				.attr("width", plot_width);
			areas_update.select("text").each(function(d) {
				const font_family = instance.fontFamily();
				const max_width_in_px = remToPx(state.y_label_max_width);
				const y_coordinate = d.range[0] + margin_top;
				const font_style = `${state.y_font_weight} ${state.y_font_size} ${font_family}`;

				const wrapped_lines_array = wrapStringToLines(d.label,
					font_style,
					null,
					max_width_in_px);


				drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, bg_class);

				drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, fg_class);

				const num_of_lines = wrapped_lines_array.length;

				const { transform_str, text_anchor } = getTransformAndAnchorForY(this, yScale, plot_width, num_of_lines, y_coordinate);

				select(this)
					.attr("transform", transform_str)
					.attr("text-anchor", text_anchor);
			})
			.attr("font-size", getYTextSize());

			areas_update.selectAll(`.${fg_class}`)
				.attr("fill", state.y_label_color);

			areas_update.selectAll(`.${bg_class}`)
				.attr("display", state.y_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.y_shadow_color)
				.attr("stroke-width", 2);


			areas.exit().remove();
		}


		function drawWrappedLabels(wrapped_lines_array, target_el, font_size, font_weight, class_name) {
			var font_size_in_px = remToPx(font_size);
			var tspans = select(target_el).selectAll(`.${class_name}`).data(wrapped_lines_array);


			tspans.join("tspan")
					.text(function (d) { return d ; })
					.attr("class", class_name)
					.attr("y", function(_, i) {
						var y = (i * (font_size_in_px * line_height));
						return y + "px";
					})
					.attr("x", 0)
					.attr("font-weight", font_weight);

			tspans.exit().remove();
		}

		function getTransformAndAnchorForX(target_el, xScale, chart_height, num_of_lines, x_coordinate_of_highlight) {
			/* the reason we are working out the text anchor and the transform at
			the same time is because we only want to calculate if a label overlaps
			the chart once, and that affects both properties*/
			let transform_str;
			let x_position_of_text = 0;
			let text_anchor;

			const x_position_middle = state.x_label_x_position == "middle";
			const y_position_bottom = state.x_label_y_position == "bottom";
			const is_vertical = state.x_label_vertical;
			const line_height_in_px = getXTextSize() * line_height;
			// svg draws from the bottom of the first line of text, so we need
			// to times line height by num of lines minus one
			const text_block_offset = line_height_in_px * (num_of_lines - 1);

			const x_padding = getXTextSize() /2;
			const y_padding_if_horizontal = y_position_bottom ? text_block_offset + x_padding: line_height_in_px;
			const y_padding_if_vertical = x_padding;
			const y_padding = is_vertical ? y_padding_if_vertical : y_padding_if_horizontal;
			const vertical_center =  is_vertical ? (chart_height / 2) : (chart_height / 2) - (text_block_offset / 2);
			const max_x_value = xScale.range()[1];
			const min_x_value = xScale.range()[0];
			const text_width = target_el.getBoundingClientRect().width;

			const y_position_lookup = {
				top: y_padding,
				middle: vertical_center,
				bottom: chart_height - y_padding
			};
			const y_position = y_position_lookup[state.x_label_y_position];

			if (is_vertical) {
				const x_position_lookup = {
					right: getXTextSize(),
					middle: - (text_block_offset - x_padding)/2,
					left: - text_block_offset - x_padding
				};
				x_position_of_text = x_position_lookup[state.x_label_x_position];

				const text_anchor_lookup = {
					top: "end",
					middle: "middle",
					bottom: "start"
				};
				text_anchor = text_anchor_lookup[state.x_label_y_position];

				const overlap_at_end = x_coordinate_of_highlight + text_width + x_padding > max_x_value;
				const overlap_at_start = x_coordinate_of_highlight - text_width - x_padding < min_x_value;

				if (overlap_at_end) x_position_of_text = x_position_lookup.left;
				if (overlap_at_start) x_position_of_text = x_position_lookup.right;

				if (x_position_middle) {
					const overlap_at_end = x_coordinate_of_highlight + (text_width /2) > max_x_value;
					const overlap_at_start = x_coordinate_of_highlight - (text_width/2) < min_x_value;

					if (overlap_at_end) x_position_of_text = x_position_lookup.left;
					if (overlap_at_start) x_position_of_text = x_position_lookup.right;
				}

				transform_str = "translate(" + x_position_of_text + ", " + y_position + ") rotate(-90)";
			}
			else if (!is_vertical) {
				// Add padding depending on which side of the line the text is on
				// If there is not enough space for the label, it will move it to
				// the other side.

				const x_position_lookup = {
					right: x_padding,
					middle: 0,
					left: - x_padding
				};
				x_position_of_text = x_position_lookup[state.x_label_x_position];

				const text_anchor_lookup = {
					right: "start",
					middle: "middle",
					left: "end"
				};
				text_anchor = text_anchor_lookup[state.x_label_x_position];

				let overlap_at_end;
				let overlap_at_start;

				if (x_position_middle) {
					overlap_at_end = x_coordinate_of_highlight + (text_width /2) > max_x_value;
					overlap_at_start = x_coordinate_of_highlight - (text_width/2) < min_x_value;
				}
				else {
					overlap_at_end = x_coordinate_of_highlight + text_width + x_padding > max_x_value;
					overlap_at_start = x_coordinate_of_highlight - text_width - x_padding < min_x_value;
				}

				if (overlap_at_end) {
					x_position_of_text = x_position_lookup.left;
					text_anchor = text_anchor_lookup.left;
				}
				if (overlap_at_start) {
					x_position_of_text = x_position_lookup.right;
					text_anchor = text_anchor_lookup.right;
				}

				transform_str = "translate(" + x_position_of_text + ", " + y_position + ")";
			}

			return {
				transform_str,
				text_anchor
			};
		}


		function getTransformAndAnchorForY(target_el, yScale, chart_width, num_of_lines, y_coordinate_of_line) {
			let text_anchor;
			let transform_str;

			const text_anchor_lookup = {
				left: "start",
				middle: "middle",
				right: "end"
			};
			text_anchor = text_anchor_lookup[state.y_label_x_position];

			const line_height_in_px = getYTextSize() * line_height;
			const text_block_offset = line_height_in_px * (num_of_lines - 1);
			const x_padding = getYTextSize()/2;
			const text_offset_and_padding = text_block_offset + x_padding;
			const max_y_val = yScale.range()[0];
			const min_y_val = yScale.range()[1];
			const text_height = target_el.getBoundingClientRect().height;
			const y_position_middle = state.y_label_y_position == "middle";

			const y_position_lookup = {
				above: - text_offset_and_padding,
				middle: - (text_offset_and_padding/2) + x_padding,
				below: line_height_in_px
			};
			let y_position = y_position_lookup[state.y_label_y_position];

			const x_position_lookup = {
				right: chart_width - x_padding,
				middle: chart_width / 2,
				left: x_padding,
			};
			let x_position = x_position_lookup[state.y_label_x_position];

			let overlap_at_top;
			let overlap_at_bottom;

			if (y_position_middle) {
				overlap_at_top = y_coordinate_of_line - (text_height/2) < min_y_val;
				overlap_at_bottom = y_coordinate_of_line + (text_height/2) > max_y_val;
			}
			else {
				overlap_at_top = y_coordinate_of_line - text_height < min_y_val;
				overlap_at_bottom = y_coordinate_of_line + text_height > max_y_val;
			}

			if (overlap_at_top) y_position = y_position_lookup.below;
			if (overlap_at_bottom) y_position = y_position_lookup.above;

			transform_str = "translate(" + x_position + ", " + y_position + ")";

			return {
				transform_str,
				text_anchor
			};
		}


		function updateHighlights() {
			var parent_element = instance.appendTo();
			var axis_highlights = parent_element.selectAll("g.axis-highlights").data(data);
			var axis_highlights_enter = axis_highlights.enter().append("g").attr("class", "axis-highlights")
				.style("opacity", 0);
			axis_highlights.merge(axis_highlights_enter)
				.transition().duration(dur)
				.style("opacity", 1)
				.each(function(d) {
					var facet_group = this;
					updateXAreas(facet_group, d);
					updateYAreas(facet_group, d);
					updateXLines(facet_group, d);
					updateYLines(facet_group, d);
				});

			axis_highlights.exit().remove();
		}

		updateHighlights();
	};
}


export { initUpdate };
